@@ -27,15 +27,20 @@ G_BEGIN_DECLS
 struct _GtkSourceLanguagePrivate 
 {
 	gchar			*lang_file_name;
+	/* this is allocated by libxml, it should be freed using xmlFree() */
 	gchar                   *translation_domain;
 
+	gchar			*id;
+	
+	/* this is allocated by libxml, it should be freed using xmlFree() */
 	gchar			*name;
+	/* this is allocated by libxml, it should be freed using xmlFree() */
 	gchar			*section;
 
 	GSList			*mime_types;
 
-	GHashTable		*tag_name_to_style_name;
-	GHashTable		*tag_name_to_style;
+	GHashTable		*tag_id_to_style_name;
+	GHashTable		*tag_id_to_style;
 
 	GtkSourceStyleScheme 	*style_scheme;
 
