@@ -21,11 +21,19 @@
 #include <config.h>
 #endif
 
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif
+
 #include <string.h>
+#include <fcntl.h>
 
 #include <libxml/xmlreader.h>
 
-#include <glib/gi18n.h>
+#include <glib/gstdio.h>
+#include <glib/gmappedfile.h>
+
+#include "gtksourceview-i18n.h"
 
 #include "gtksourcelanguage-private.h"
 #include "gtksourcelanguage.h"
@@ -66,13 +74,19 @@ _gtk_source_language_new_from_file (const gchar			*filename,
 {
 	GtkSourceLanguage *lang = NULL;
 
-	xmlTextReaderPtr reader;
+	xmlTextReaderPtr reader = NULL;
 	gint ret;
-	
+	int fd;
+
 	g_return_val_if_fail (filename != NULL, NULL);
 	g_return_val_if_fail (lm != NULL, NULL);
 
-	reader = xmlNewTextReaderFilename (filename);
+	/*
+	 * Use fd instead of filename so that it's utf8 safe on w32.
+	 */
+	fd = g_open (filename, O_RDONLY, 0);
+	if (fd != -1)
+		reader = xmlReaderForFd (fd, filename, NULL, 0);
 
 	if (reader != NULL) 
 	{
@@ -86,7 +100,7 @@ _gtk_source_language_new_from_file (const gchar			*filename,
 
 				name = xmlTextReaderName (reader);
 
-				if (strcmp (name, "language") == 0)
+				if (xmlStrcmp (name, BAD_CAST "language") == 0)
 				{
 
 					lang = process_language_node (reader, filename);
@@ -103,7 +117,8 @@ _gtk_source_language_new_from_file (const gchar			*filename,
 		}
 	
 		xmlFreeTextReader (reader);
-        	
+		close (fd);
+
 		if (ret != 0) 
 		{
 	            g_warning("Failed to parse '%s'", filename);
@@ -119,7 +134,6 @@ _gtk_source_language_new_from_file (const gchar			*filename,
 	return lang;
 }
 
-
 GType
 gtk_source_language_get_type (void)
 {
@@ -264,8 +278,8 @@ escape_id (const gchar *arbitrary_text, gint len)
 static GtkSourceLanguage *
 process_language_node (xmlTextReaderPtr reader, const gchar *filename)
 {
-	gchar *version;
-	gchar *mimetypes;
+	xmlChar *version;
+	xmlChar *mimetypes;
 	gchar** mtl;
 	int i;
 	xmlChar *tmp;
@@ -276,16 +290,21 @@ process_language_node (xmlTextReaderPtr reader, const gchar *filename)
 
 	lang->priv->lang_file_name = g_strdup (filename);
 	
-	lang->priv->translation_domain = xmlTextReaderGetAttribute (reader, "translation-domain");
+	lang->priv->translation_domain = (gchar *) xmlTextReaderGetAttribute (
+			reader, BAD_CAST "translation-domain");
 	if (lang->priv->translation_domain == NULL)
 	{
-		lang->priv->translation_domain = xmlStrdup (GETTEXT_PACKAGE);
+		/* if the attribute "translation-domain" exists then
+		 * lang->priv->translation_domain is a xmlChar so it must always
+		 * be a xmlChar, this is why xmlStrdup() is used instead of
+		 * g_strdup() */
+		lang->priv->translation_domain = (gchar *)xmlStrdup (BAD_CAST GETTEXT_PACKAGE);
 	}
 	
-	tmp = xmlTextReaderGetAttribute (reader, "_name");
+	tmp = xmlTextReaderGetAttribute (reader, BAD_CAST "_name");
 	if (tmp == NULL)
 	{
-		lang->priv->name = xmlTextReaderGetAttribute (reader, "name");
+		lang->priv->name = (gchar *)xmlTextReaderGetAttribute (reader, BAD_CAST "name");
 		if (lang->priv->name == NULL)
 		{
 			g_warning ("Impossible to get language name from file '%s'",
@@ -295,23 +314,28 @@ process_language_node (xmlTextReaderPtr reader, const gchar *filename)
 			return NULL;
 		}
 
-		id_temp = xmlStrdup (lang->priv->name);
+		id_temp = xmlStrdup (BAD_CAST lang->priv->name);
 	}
 	else
 	{
 		id_temp = xmlStrdup (tmp);
-		lang->priv->name = xmlStrdup (dgettext (lang->priv->translation_domain, tmp));
+		/* if tmp is NULL then lang->priv->name is a xmlChar so it must
+		 * always be a xmlChar, this is why xmlStrdup() is used instead
+		 * of g_strdup() */
+		lang->priv->name = (gchar *)xmlStrdup (BAD_CAST dgettext (
+					lang->priv->translation_domain,
+					(gchar *)tmp));
 		xmlFree (tmp);
 	}
 
 	g_return_val_if_fail (id_temp != NULL, NULL);
-	lang->priv->id = escape_id (id_temp, -1);
+	lang->priv->id = escape_id ((gchar *)id_temp, -1);
 	xmlFree (id_temp);
 
-	tmp = xmlTextReaderGetAttribute (reader, "_section");
+	tmp = xmlTextReaderGetAttribute (reader, BAD_CAST "_section");
 	if (tmp == NULL)
 	{
-		lang->priv->section = xmlTextReaderGetAttribute (reader, "section");
+		lang->priv->section = (gchar *)xmlTextReaderGetAttribute (reader, BAD_CAST "section");
 		if (lang->priv->section == NULL)
 		{
 			g_warning ("Impossible to get language section from file '%s'",
@@ -323,11 +347,16 @@ process_language_node (xmlTextReaderPtr reader, const gchar *filename)
 	}
 	else
 	{
-		lang->priv->section = xmlStrdup (dgettext (lang->priv->translation_domain, tmp));
+		/* if tmp is NULL then lang->priv->section is a xmlChar so it
+		 * must always be a xmlChar, this is why xmlStrdup() is used
+		 * instead of g_strdup() */
+		lang->priv->section = (gchar *)xmlStrdup (BAD_CAST dgettext (
+					lang->priv->translation_domain,
+					(gchar *)tmp));
 		xmlFree (tmp);
 	}
 	
-	version = xmlTextReaderGetAttribute (reader, "version");
+	version = xmlTextReaderGetAttribute (reader, BAD_CAST "version");
 	if (version == NULL)
 	{
 		g_warning ("Impossible to get version number from file '%s'",
@@ -338,10 +367,10 @@ process_language_node (xmlTextReaderPtr reader, const gchar *filename)
 	}
 	else
 	{
-		if (strcmp (version , "1.0") != 0)
+		if (xmlStrcmp (version , BAD_CAST "1.0") != 0)
 		{
 			g_warning ("Usupported language spec version '%s' in file '%s'",
-				   version, filename);
+				   (gchar *)version, filename);
 
 			xmlFree (version);
 			
@@ -352,7 +381,7 @@ process_language_node (xmlTextReaderPtr reader, const gchar *filename)
 		xmlFree (version);
 	}
 
-	mimetypes = xmlTextReaderGetAttribute (reader, "mimetypes");
+	mimetypes = xmlTextReaderGetAttribute (reader, BAD_CAST "mimetypes");
 	if (mimetypes == NULL)
 	{
 		g_warning ("Impossible to get mimetypes from file '%s'",
@@ -362,19 +391,16 @@ process_language_node (xmlTextReaderPtr reader, const gchar *filename)
 		return NULL;
 	}
 
-	mtl = g_strsplit (mimetypes, ";" , 0);
+	mtl = g_strsplit ((gchar *)mimetypes, ";", 0);
 
-	i = 0; 
-	
-	do
+	for (i = 0; mtl[i] != NULL; i++)
 	{
+		/* steal the strings from the array */
 		lang->priv->mime_types = g_slist_prepend (lang->priv->mime_types,
-				g_strdup (mtl[i]));
+							  mtl[i]);
+	}
 
-		++i;
-	} while (mtl[i] != NULL);
-
-	g_strfreev (mtl);
+	g_free (mtl);
 	xmlFree (mimetypes);
 
 	lang->priv->mime_types = g_slist_reverse (lang->priv->mime_types);
@@ -471,14 +497,20 @@ gtk_source_language_get_mime_types (GtkSourceLanguage *language)
 static GSList *
 get_mime_types_from_file (GtkSourceLanguage *language)
 {
-	xmlTextReaderPtr reader;
+	xmlTextReaderPtr reader = NULL;
 	gint ret;
 	GSList *mime_types = NULL;
+	int fd;
 
 	g_return_val_if_fail (GTK_IS_SOURCE_LANGUAGE (language), NULL);
 	g_return_val_if_fail (language->priv->lang_file_name != NULL, NULL);
 		
-	reader = xmlNewTextReaderFilename (language->priv->lang_file_name);
+	/*
+	 * Use fd instead of filename so that it's utf8 safe on w32.
+	 */
+	fd = g_open (language->priv->lang_file_name, O_RDONLY, 0);
+	if (fd != -1)
+		reader = xmlReaderForFd (fd, language->priv->lang_file_name, NULL, 0);
 
 	if (reader != NULL) 
 	{
@@ -492,13 +524,12 @@ get_mime_types_from_file (GtkSourceLanguage *language)
 
 				name = xmlTextReaderName (reader);
 
-				if (strcmp (name, "language") == 0)
+				if (xmlStrcmp (name, BAD_CAST "language") == 0)
 				{
 					gchar *mimetypes;
-					gchar** mtl;
-					gint i;
 
-					mimetypes = xmlTextReaderGetAttribute (reader, "mimetypes");
+					mimetypes = (gchar *)xmlTextReaderGetAttribute (reader,
+							BAD_CAST "mimetypes");
 					
 					if (mimetypes == NULL)
 					{
@@ -509,20 +540,19 @@ get_mime_types_from_file (GtkSourceLanguage *language)
 					}
 					else
 					{
+						gchar **mtl;
+						gint i;
 
-						mtl = g_strsplit (mimetypes, ";" , 0);
+						mtl = g_strsplit (mimetypes, ";", 0);
 
-						i = 0; 
-	
-						do
+						for (i = 0; mtl[i] != NULL; i++)
 						{
+							/* steal the strings from the array */
 							mime_types = g_slist_prepend (mime_types,
-										      g_strdup (mtl[i]));
+										      mtl[i]);
+						}
 
-							++i;
-						} while (mtl[i] != NULL);
-
-						g_strfreev (mtl);
+						g_free (mtl);
 						xmlFree (mimetypes);
 
 						ret = 0;
@@ -534,11 +564,11 @@ get_mime_types_from_file (GtkSourceLanguage *language)
 			
 			if (ret != 0)
 				ret = xmlTextReaderRead (reader);
-			
 		}
 	
 		xmlFreeTextReader (reader);
-        	
+		close (fd);
+
 		if (ret != 0) 
 		{
 	            g_warning("Failed to parse '%s'", language->priv->lang_file_name);
@@ -642,7 +672,7 @@ parseLineComment (xmlDocPtr doc, xmlNodePtr cur, gchar* id, xmlChar *name)
 			
 		start_regex = xmlNodeListGetString (doc, child->xmlChildrenNode, 1);
 			
-		tag = gtk_line_comment_tag_new (id, name, strconvescape (start_regex));
+		tag = gtk_line_comment_tag_new (id, (gchar *)name, strconvescape ((gchar *)start_regex));
 
 		xmlFree (start_regex);
 	}
@@ -701,9 +731,9 @@ parseBlockComment (xmlDocPtr doc, xmlNodePtr cur, gchar* id, xmlChar *name)
 	}
 
 	tag = gtk_block_comment_tag_new (id,
-					 name, 
-					 strconvescape (start_regex), 
-					 strconvescape (end_regex));
+					 (gchar *)name, 
+					 strconvescape ((gchar *)start_regex), 
+					 strconvescape ((gchar *)end_regex));
 
 	xmlFree (start_regex);
 	xmlFree (end_regex);
@@ -724,11 +754,11 @@ parseString (xmlDocPtr doc, xmlNodePtr cur, gchar* id, xmlChar *name)
 
 	xmlNodePtr child;
 
-	prop = xmlGetProp (cur, "end-at-line-end");
+	prop = xmlGetProp (cur, BAD_CAST "end-at-line-end");
 	if (prop != NULL)
 	{
-		if (!xmlStrcasecmp (prop, (const xmlChar *)"TRUE") ||
-		    !xmlStrcmp (prop, (const xmlChar *)"1"))
+		if (!xmlStrcasecmp (prop, BAD_CAST "TRUE") ||
+		    !xmlStrcmp (prop, BAD_CAST "1"))
 
 				end_at_line_end = TRUE;
 			else
@@ -773,9 +803,9 @@ parseString (xmlDocPtr doc, xmlNodePtr cur, gchar* id, xmlChar *name)
 	}
 
 	tag = gtk_string_tag_new (id,
-				  name, 
-				  strconvescape (start_regex), 
-				  strconvescape (end_regex), 
+				  (gchar *)name, 
+				  strconvescape ((gchar *)start_regex), 
+				  strconvescape ((gchar *)end_regex), 
 				  end_at_line_end);
 
 	xmlFree (start_regex);
@@ -801,7 +831,7 @@ parseKeywordList (xmlDocPtr doc, xmlNodePtr cur, gchar* id, xmlChar *name)
 
 	xmlNodePtr child;
 
-	prop = xmlGetProp (cur, "case-sensitive");
+	prop = xmlGetProp (cur, BAD_CAST "case-sensitive");
 	if (prop != NULL)
 	{
 		if (!xmlStrcasecmp (prop, (const xmlChar *)"TRUE") ||
@@ -814,7 +844,7 @@ parseKeywordList (xmlDocPtr doc, xmlNodePtr cur, gchar* id, xmlChar *name)
 		xmlFree (prop);	
 	}
 
-	prop = xmlGetProp (cur, "match-empty-string-at-beginning");
+	prop = xmlGetProp (cur, BAD_CAST "match-empty-string-at-beginning");
 	if (prop != NULL)
 	{
 		if (!xmlStrcasecmp (prop, (const xmlChar *)"TRUE") ||
@@ -827,7 +857,7 @@ parseKeywordList (xmlDocPtr doc, xmlNodePtr cur, gchar* id, xmlChar *name)
 		xmlFree (prop);	
 	}
 
-	prop = xmlGetProp (cur, "match-empty-string-at-end");
+	prop = xmlGetProp (cur, BAD_CAST "match-empty-string-at-end");
 	if (prop != NULL)
 	{
 		if (!xmlStrcasecmp (prop, (const xmlChar *)"TRUE") ||
@@ -840,18 +870,18 @@ parseKeywordList (xmlDocPtr doc, xmlNodePtr cur, gchar* id, xmlChar *name)
 		xmlFree (prop);	
 	}
 
-	prop = xmlGetProp (cur, "beginning-regex");
+	prop = xmlGetProp (cur, BAD_CAST "beginning-regex");
 	if (prop != NULL)
 	{
-		beginning_regex = g_strdup (prop);
+		beginning_regex = g_strdup ((gchar *)prop);
 		
 		xmlFree (prop);	
 	}
 
-	prop = xmlGetProp (cur, "end-regex");
+	prop = xmlGetProp (cur, BAD_CAST "end-regex");
 	if (prop != NULL)
 	{
-		end_regex = g_strdup (prop);
+		end_regex = g_strdup ((gchar *)prop);
 		
 		xmlFree (prop);	
 	}
@@ -865,7 +895,7 @@ parseKeywordList (xmlDocPtr doc, xmlNodePtr cur, gchar* id, xmlChar *name)
 			xmlChar *keyword;
 			keyword = xmlNodeListGetString (doc, child->xmlChildrenNode, 1);
 			
-			list = g_slist_prepend (list, strconvescape (keyword));
+			list = g_slist_prepend (list, strconvescape ((gchar *)keyword));
 		}
 
 		child = child->next;
@@ -885,7 +915,7 @@ parseKeywordList (xmlDocPtr doc, xmlNodePtr cur, gchar* id, xmlChar *name)
 	}
 
 	tag = gtk_keyword_list_tag_new (id,
-					name, 
+					(gchar *)name, 
 					list,
 					case_sensitive,
 					match_empty_string_at_beginning,
@@ -917,7 +947,8 @@ parsePatternItem (xmlDocPtr doc, xmlNodePtr cur, gchar* id, xmlChar *name)
 			
 		regex = xmlNodeListGetString (doc, child->xmlChildrenNode, 1);
 			
-		tag = gtk_pattern_tag_new (id, name, strconvescape (regex));
+		tag = gtk_pattern_tag_new (id, (gchar *)name,
+				strconvescape ((gchar *)regex));
 
 		xmlFree (regex);
 	}
@@ -976,9 +1007,9 @@ parseSyntaxItem (xmlDocPtr doc, xmlNodePtr cur, gchar* id, xmlChar *name)
 	}
 
 	tag = gtk_syntax_tag_new (id,
-				  name, 
-				  strconvescape (start_regex),
-				  strconvescape (end_regex));
+				  (gchar *)name, 
+				  strconvescape ((gchar *)start_regex),
+				  strconvescape ((gchar *)end_regex));
 
 	xmlFree (start_regex);
 	xmlFree (end_regex);
@@ -1023,24 +1054,26 @@ parseTag (GtkSourceLanguage *language,
 	GtkTextTag *tag = NULL;
 	xmlChar *name;
 	xmlChar *style;
-	xmlChar *id_temp;
+	gchar   *id_temp;
 	gchar   *id;
 	
-	name = xmlGetProp (cur, "_name");
+	name = xmlGetProp (cur, BAD_CAST "_name");
 	if (name == NULL)
 	{
-		name = xmlGetProp (cur, "name");
-		id_temp =xmlStrdup (name);
+		name = xmlGetProp (cur, BAD_CAST "name");
+		id_temp = g_strdup ((gchar *)name);
 	}
 	else
 	{
-		xmlChar *tmp = xmlStrdup (dgettext (language->priv->translation_domain, name));
-		id_temp = xmlStrdup (name);
+		xmlChar *tmp = xmlStrdup (BAD_CAST dgettext (
+					language->priv->translation_domain,
+					(gchar *)name));
+		id_temp = g_strdup ((gchar *)name);
 		xmlFree (name);
 		name = tmp;
 	}
 	
-	style = xmlGetProp (cur, "style");
+	style = xmlGetProp (cur, BAD_CAST "style");
 
 	if (name == NULL)
 	{
@@ -1049,12 +1082,12 @@ parseTag (GtkSourceLanguage *language,
 
 	g_return_val_if_fail (id_temp != NULL, tag_list);
 	id = escape_id (id_temp, -1);
-	xmlFree (id_temp);
+	g_free (id_temp);
 
 	if (style == NULL)
 	{
 		/* FIXME */
-		style = xmlStrdup ("Normal");
+		style = xmlStrdup (BAD_CAST "Normal");
 	}
 	
 	if (!xmlStrcmp (cur->name, (const xmlChar *)"line-comment"))
@@ -1095,7 +1128,7 @@ parseTag (GtkSourceLanguage *language,
 		if (populate_styles_table)
 			g_hash_table_insert (language->priv->tag_id_to_style_name, 
 					     g_strdup (id), 
-					     g_strdup (style));
+					     g_strdup ((gchar *)style));
 
 		ts = gtk_source_language_get_tag_style (language, id);
 
@@ -1127,10 +1160,22 @@ language_file_parse (GtkSourceLanguage *language,
 	GSList *tag_list = NULL;
 	xmlDocPtr doc;
 	xmlNodePtr cur;
-
+	GMappedFile *mf;
+	
 	xmlKeepBlanksDefault (0);
 
-	doc = xmlParseFile (language->priv->lang_file_name);
+	mf = g_mapped_file_new (language->priv->lang_file_name, FALSE, NULL);
+	
+	if (mf == NULL)
+		doc = NULL;
+	else
+	{
+		doc = xmlParseMemory (g_mapped_file_get_contents (mf), 
+				      g_mapped_file_get_length (mf));
+
+		g_mapped_file_free (mf);
+	}
+	
 	if (doc == NULL)
 	{
 		g_warning ("Impossible to parse file '%s'",
@@ -1172,7 +1217,7 @@ language_file_parse (GtkSourceLanguage *language,
 			gunichar esc_char;
 			
 			escape = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);
-			esc_char = g_utf8_get_char_validated (escape, -1);
+			esc_char = g_utf8_get_char_validated ((gchar *)escape, -1);
 			if (esc_char < 0)
 			{
 				g_warning ("Invalid (non UTF8) escape character in file '%s'",
