@@ -21,6 +21,11 @@
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
+/* If TEST_XML_MEM is defined the test program will try to detect memory
+ * allocated by xmlMalloc() but not freed by xmlFree() or freed by xmlFree()
+ * but not allocated by xmlMalloc(). */
+/*#define TEST_XML_MEM*/
+
 #include <stdio.h>
 #include <string.h>
 #include <gtk/gtk.h>
@@ -32,7 +37,9 @@
 #include <gtksourceview/gtksourcelanguage.h>
 #include <gtksourceview/gtksourcelanguagesmanager.h>
 #include <gtksourceview/gtksourceprintjob.h>
-
+#ifdef TEST_XML_MEM
+#include <libxml/xmlreader.h>
+#endif
 
 /* Global list of open windows */
 
@@ -60,6 +67,8 @@ static void       markers_toggled_cb             (GtkAction       *action,
 						  gpointer         user_data);
 static void       margin_toggled_cb              (GtkAction       *action,
 						  gpointer         user_data);
+static void       hl_line_toggled_cb             (GtkAction       *action,
+						  gpointer         user_data);
 static void       auto_indent_toggled_cb         (GtkAction       *action,
 						  gpointer         user_data);
 static void       insert_spaces_toggled_cb       (GtkAction       *action,
@@ -103,6 +112,9 @@ static GtkToggleActionEntry toggle_entries[] = {
 	{ "ShowMargin", NULL, "Show M_argin", NULL,
 	  "Toggle visibility of right margin indicator",
 	  G_CALLBACK (margin_toggled_cb), FALSE },
+	{ "HlLine", NULL, "_Highlight Current Line", NULL,
+	  "Toggle highlighting of current line",
+	  G_CALLBACK (hl_line_toggled_cb), FALSE },
 	{ "AutoIndent", NULL, "Enable _Auto Indent", NULL,
 	  "Toggle automatic auto indentation of text",
 	  G_CALLBACK (auto_indent_toggled_cb), FALSE },
@@ -131,6 +143,7 @@ static const gchar *view_ui_description =
 "      <menuitem action=\"ShowNumbers\"/>"
 "      <menuitem action=\"ShowMarkers\"/>"
 "      <menuitem action=\"ShowMargin\"/>"
+"      <menuitem action=\"HlLine\"/>"
 "      <separator/>"
 "      <menuitem action=\"AutoIndent\"/>"
 "      <menuitem action=\"InsertSpaces\"/>"
@@ -384,7 +397,8 @@ finished_cb (GtkSourcePrintJob *job, gpointer user_data)
 
 	g_print ("\n");
 	gjob = gtk_source_print_job_get_print_job (job);
-	preview = gnome_print_job_preview_new (gjob, "test-widget print preview");
+	preview = gnome_print_job_preview_new (gjob,
+		(const guchar *)"test-widget print preview");
  	g_object_unref (gjob); 
  	g_object_unref (job);
 	
@@ -421,6 +435,16 @@ margin_toggled_cb (GtkAction *action, gpointer user_data)
 		gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action)));
 }
 
+static void
+hl_line_toggled_cb (GtkAction *action, gpointer user_data)
+{
+	g_return_if_fail (GTK_IS_TOGGLE_ACTION (action) && GTK_IS_SOURCE_VIEW (user_data));
+	gtk_source_view_set_highlight_current_line (
+		GTK_SOURCE_VIEW (user_data),
+		gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action)));
+}
+
+
 static void 
 auto_indent_toggled_cb (GtkAction *action,
 			gpointer   user_data)
@@ -828,6 +852,10 @@ create_view_window (GtkSourceBuffer *buffer, GtkSourceView *from)
 		gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
 					      gtk_source_view_get_show_margin (from));
 		
+		action = gtk_action_group_get_action (action_group, "HlLine");
+		gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
+					      gtk_source_view_get_highlight_current_line (from));
+
 		action = gtk_action_group_get_action (action_group, "AutoIndent");
 		gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
 					      gtk_source_view_get_auto_indent (from));
@@ -918,7 +946,7 @@ create_main_window (GtkSourceBuffer *buffer)
 	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), TRUE);
 
 	action = gtk_action_group_get_action (action_group, "ShowMargin");
-	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), TRUE);
+	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), FALSE);
 
 	action = gtk_action_group_get_action (action_group, "AutoIndent");
 	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), TRUE);
@@ -949,6 +977,87 @@ create_source_buffer (GtkSourceLanguagesManager *manager)
 }
 
 
+/* XML memory management and verification functions ----------------------------- */
+
+#ifdef TEST_XML_MEM
+
+static GHashTable *xml_mem_table = NULL;
+
+static void
+xml_free (void *mem)
+{
+	if (mem == NULL)
+		return;
+
+	if (g_hash_table_remove (xml_mem_table, mem))
+	{
+		g_free (mem);
+	}
+	else
+	{
+		g_warning ("Memory at %p (\"%s\") was not allocated by libxml",
+				mem, (gchar*)mem);
+	}
+}
+
+static void *
+xml_malloc (size_t size)
+{
+	void *allocated_mem = g_malloc (size);
+	g_hash_table_insert (xml_mem_table, allocated_mem, GINT_TO_POINTER (TRUE));
+	return allocated_mem;
+}
+
+static void *
+xml_realloc (void *mem, size_t size)
+{
+	void *allocated_mem;
+
+	if (!g_hash_table_remove (xml_mem_table, mem))
+	{
+		g_warning ("Memory at %p (\"%s\") was not allocated by libxml",
+				mem, (gchar*)mem);
+	}
+
+	allocated_mem = g_realloc (mem, size);
+	g_hash_table_insert (xml_mem_table, allocated_mem, GINT_TO_POINTER (TRUE));
+	return allocated_mem;
+}
+
+static char *
+xml_strdup (const char *str)
+{
+	void *allocated_mem = g_strdup (str);
+	g_hash_table_insert (xml_mem_table, allocated_mem, GINT_TO_POINTER (TRUE));
+	return allocated_mem;
+}
+
+static void
+xml_init ()
+{
+	xml_mem_table = g_hash_table_new (NULL, NULL);
+	if (xmlMemSetup (xml_free, xml_malloc, xml_realloc, xml_strdup) != 0)
+		g_warning ("xmlMemSetup() failed");
+}
+
+static void
+xml_table_foreach_cb (gpointer key, gpointer value, gpointer user_data)
+{
+	/* Some of this memory could be internally allocated by libxml. */
+	g_warning ("Memory at %p (\"%s\") was not freed, freed without using xmlFree() "
+			"or allocated internally by libxml", key, (gchar*)key);
+}
+
+static void
+xml_finalize ()
+{
+	g_hash_table_foreach (xml_mem_table, xml_table_foreach_cb, NULL);
+	g_hash_table_destroy (xml_mem_table);
+}
+
+#endif /* TEST_XML_MEM */
+
+
 /* Program entry point ------------------------------------------------------------ */
 
 int
@@ -961,7 +1070,11 @@ main (int argc, char *argv[])
 	/* initialization */
 	gtk_init (&argc, &argv);
 	gnome_vfs_init ();
-	
+
+#ifdef TEST_XML_MEM
+	xml_init ();
+#endif
+
 	/* create buffer */
 	lm = gtk_source_languages_manager_new ();
 	buffer = create_source_buffer (lm);
@@ -986,6 +1099,10 @@ main (int argc, char *argv[])
 	g_object_unref (buffer);
 
 	gnome_vfs_shutdown ();
-	
+
+#ifdef TEST_XML_MEM
+	xml_finalize ();
+#endif
+
 	return 0;
 }
